//----------------------------------------------------------- 2013-09-29 -----------
// 電力計　無線中継プログラム Ver 1.0.1
//--------------------------------------------- written by iizuka   Japan ----------
#include <SPI.h>
#include <Mirf.h>
#include <MirfHardwareSpiDriver.h>

#define PORT_BUZZER     19  // Port Analog   5

byte radio_buf_read[32];     // 無線受信バッファ
byte radio_buf_write[32];    // 無線送信バッファ
byte serial_buf_read[32];    // シリアル通信受信バッファ
byte serial_buf_index;       // シリアル通信受信バッファインデックス

bool pc_ready = false;       // PCが受信可能な状態にある
unsigned long watchdog_2k;   // Windowsプログラムが正常動作しているか確認するためのタイマー
unsigned long watchdog_5k;   // Windowsプログラムが正常動作しているか確認するためのタイマー

//**********debug****************************
unsigned long seq_tmp = 1;
typedef union {
  unsigned long long_val;
  struct {
    byte LL;  // Lower address
    byte ML;
    byte MH;
    byte HH;  // Higher address
  } byte_val;    
} type_convert_long_byte; 
type_convert_long_byte  v;
type_convert_long_byte  w;
//**************************************

typedef union {
  unsigned int int_val;
  struct {
    byte L;  // Lower address
    byte H;
  } byte_val;    
} type_convert_int_byte;   
type_convert_int_byte btoi;
  
//------------------------------------------------------------------------------------------
// タイムアウト付き無線受信
//
// timeout: タイムアウト時間[msec]
//------------------------------------------------------------------------------------------
bool radio_recieve( byte buf[], unsigned long timeout )
{
  unsigned long millis_tmp;
  bool millis_overflow = false;
  
  unsigned long started_waiting_at = millis();
  while(!Mirf.dataReady()){
    if( millis_diff(started_waiting_at) > timeout ) {
      return false;
    }
  }    
  Mirf.getData( buf );
  return true;
}

//------------------------------------------------------------------------------------------
// 無線送信
//------------------------------------------------------------------------------------------
void radio_write( byte buf[] )
{
  Mirf.send( buf );
  
  while( Mirf.isSending() ){}
}

//------------------------------------------------------------------------------------------
// 引数と現在値とのミリ秒差を返す　
//   millis()の50日オーバーフローは回避されるが差が3.1日を越えると誤動作する
//------------------------------------------------------------------------------------------
unsigned long millis_diff(unsigned long millis_from)
{
  if( millis_from > 0xF0000000 ) {
    millis_from += 0x10000000;
    return (millis() + 0x10000000) - millis_from;
  } else {
    return millis() - millis_from;
  }
}

//------------------------------------------------------------------------------------------
// ブザーを０．1秒鳴らす
//------------------------------------------------------------------------------------------  
void beep()
{
  for( int i = 0 ; i < 50 ; i++ ) {
    digitalWrite( PORT_BUZZER, HIGH );
    delay(1);
    digitalWrite( PORT_BUZZER, LOW  );
    delay(1);
  }    
}

//------------------------------------------------------------------------------------------
//  arduino setup
//------------------------------------------------------------------------------------------
void setup(void)
{
  pinMode(PORT_BUZZER, OUTPUT);
    
  Serial.begin(57600);

  // 通信開始
  Mirf.csnPin = 10;
  Mirf.cePin  = 9;
  Mirf.spi = &MirfHardwareSpi;
  Mirf.init();

  Mirf.setRADDR((byte *)"clie2");
  Mirf.setTADDR((byte *)"serv2");
  Mirf.payload = 32;
  Mirf.channel = 3;
  Mirf.config();

  serial_buf_index = 0;
  seq_tmp = 0;
}

//------------------------------------------------------------------------------------------
//  arduino loop
//------------------------------------------------------------------------------------------
void loop(void)
{
  byte incomingByte;
  unsigned long get_command_at;
  byte eol_len;

  //**********  [WindowsPC] ---Serial---> [thisPGM] ---Radio---> [Client] **********
  if (Serial.available() > 0) {
    incomingByte = Serial.read();
    if( incomingByte == '#' ) {
      serial_buf_read[serial_buf_index] = '\0';
      switch(serial_buf_read[0]) {
        case 'S':   // Start
          Serial.print("s#");
          pc_ready = true;
          watchdog_2k = millis();
          watchdog_5k = millis();
          break;
        case 'W':   // Wait
          pc_ready = false;
          radio_buf_write[0] = 'W';
          radio_write( radio_buf_write );
          Serial.print("w#w#");
          break;
        case 't':  // 時刻応答
          radio_buf_write[0] = 't';
          for( int i = 1 ; i < 20 ; i++ ) {
            radio_buf_write[i] = serial_buf_read[i];
          }
          radio_write( radio_buf_write );
          break;
        case 'q':  // WatchDogs応答
          watchdog_5k = millis();
          break;
        case 'd':  
          w.long_val = atol((const char *)&serial_buf_read[1] );
          Serial.print("%REWIND ");
          Serial.print(w.long_val);
          Serial.print("#");
          seq_tmp = w.long_val;
         /*   radio_buf_write[0] = 'd';
          radio_buf_write[1] = w.byte_val.LL;
          radio_buf_write[2] = w.byte_val.ML;
          radio_buf_write[3] = w.byte_val.MH;
          radio_buf_write[4] = w.byte_val.HH;
          radio_write( radio_buf_write );
          */
      }
      serial_buf_index = 0; 
    } else {
      if( incomingByte != 0x0a ) {
        serial_buf_read[serial_buf_index] = incomingByte;
        serial_buf_index++;
      }
    }
  }

  //**********  [WindowsPC] <---Serial--- [thisPGM] <---無線--- [Client] **********
  if( radio_recieve( radio_buf_read, 10 ) ) {
    switch( radio_buf_read[0] ) {
      case 'T':  // Time request
        Serial.print("T#"); // query time
        break;
      case 'D':  // Watt data
        v.byte_val.LL = radio_buf_read[3];
        v.byte_val.ML = radio_buf_read[4]; 
        v.byte_val.MH = radio_buf_read[5]; 
        v.byte_val.HH = radio_buf_read[6];
        if( seq_tmp == 0 || seq_tmp > v.long_val ) {
          seq_tmp = v.long_val;
        }
        if( seq_tmp != v.long_val ) {
          Serial.print("%SEQ ERROR ");Serial.print(seq_tmp);Serial.print(" ");Serial.print(v.long_val);Serial.print("#");
          radio_buf_write[0] = 'd';
          w.long_val =  seq_tmp;
          radio_buf_write[1] = w.byte_val.LL;
          radio_buf_write[2] = w.byte_val.ML;
          radio_buf_write[3] = w.byte_val.MH;
          radio_buf_write[4] = w.byte_val.HH;
          radio_write( radio_buf_write );
        } else { 
          Serial.print("D");
          Serial.print(radio_buf_read[1]);
          Serial.write(radio_buf_read[2]);
          Serial.print(v.long_val);
          for( byte i = 0 ; i < radio_buf_read[1] ; i++ ) {
            btoi.byte_val.L = radio_buf_read[i*4+7];
            btoi.byte_val.H = radio_buf_read[i*4+8];
            Serial.print(",");
            Serial.print(btoi.int_val);
            btoi.byte_val.L = radio_buf_read[i*4+9];
            btoi.byte_val.H = radio_buf_read[i*4+10];
            Serial.print(",");
            Serial.print(btoi.int_val);          
          }
          if( radio_buf_read[2] == 'T' ) {
            Serial.print(",");
            Serial.write(&radio_buf_read[11],19);
          }
          Serial.print("#");
          seq_tmp += (unsigned long)radio_buf_read[1];
        }
        break;
      case 'P':  // Ping
        radio_buf_write[0] = 'p';
        if( pc_ready ) {
          radio_buf_write[1] = 'R';
        } else {
          radio_buf_write[1] = 'N';
        }
        radio_write( radio_buf_write );
        break;
      case 'w':  // Time request
        Serial.print("w#"); // Ack pause
        break;
      default:
        Serial.print("%ERROR case else#");
        break;
    }
  }
  
  //---------- WatchDogs --------------------
  if( pc_ready ) {
    if( millis_diff( watchdog_2k ) > 2000 ) {
      Serial.print("Q#"); 
      watchdog_2k = millis();
    }

    if( millis_diff( watchdog_5k ) > 5000 ) {  
      pc_ready = false;      // WindowsPCから q# 応答がある限りこの処理は行われない
                             // 一旦falseになった場合はWindowsPCからS#コマンドが送られてくるまでtrueにならない
    }
  }
}
//-------- end of program   ---------------------------------------------------------------------
